/////////////////////////////////////////////////////////////////////////
//	"udflib.cpp"
//------------------------------------------------------------------------
// Reference:
//   Linux UDF Filesystem Driver
//   UDF & ECMA 167 Documents
//     http://www.osta.org/
//     http://www.ecma.ch/
//------------------------------------------------------------------------
//	$Id: udflib.cpp 176 2006-05-26 12:05:14Z taro $
//////////////////////////////////////////////////////////////////////////

#ifdef	HAVE_CONFIG_H
#	include	"config.h"
#	undef	HAVE_CONFIG_H
#endif
#include	"udflib.h"
#include	<stdio.h>

#define E_REGID_NSR      "+NSR02"
#define E_REGID_IUVD     "*UDF LV Info"
#define E_REGID_FREE_EA  "*UDF FreeEASpace"
#define E_REGID_OS2_EA   "*OS/2 EA"
#define E_REGID_OS2_EAL  "*OS/2 EALength"
#define E_REGID_MAC_VOL  "*UDF Mac VolumeInfo"
#define E_REGID_MAC_FIND "*UDF Mac FinderInfo"
#define E_REGID_MAC_UNQ  "*UDF Mac UniqueIDTable"
#define E_REGID_MAC_RES  "*UDF Mac ResourceFork"
#define E_REGID_CD_VP    "*UDF Virtual Partition"
#define E_REGID_VAT      "*UDF Virtual Alloc Tbl"
#define E_REGID_SPARE    "*UDF Sparing Table"
#define E_REGID_CD_SP    "*UDF Sparable Partition"

UdfLib::UdfLib()
{
	this->devCtl = NULL;
	this->inqData = NULL;
	this->capa = NULL;
	this->sec_off = 0;
	this->sec_num = 0;
}

UdfLib::~UdfLib()
{
}

int		UdfLib::OpenUdf(DevCtl & devCtl, DevCtl::ScsiInqData & inq, DevCtl::SCSICAPACITYDATA16 & capa, INT64 sec_off, INT64 sec_num)
{
	this->devCtl  = &devCtl;
	this->inqData = &inq;
	this->capa    = &capa;
	this->sec_off = sec_off;
	this->sec_num = sec_num;
	if ( this->sec_off + this->sec_num > this->devCtl->ScsiGet64(this->capa->logicalBlockAddress) )
		return -1;	// error
	return 0;
}

//////////////////////////////////////////////////////////////////////////
//	Get Volume Recognition Structures

int		UdfLib::CheckVRS(UdfLib::VRSInfo & vrsInfo)
{
	BYTE						*buffer = NULL;
	INT64						blkNum;
	unsigned					blockSize;
	int							retcode;
	static struct tagVrsSid
	{
		unsigned	type;
		char		sid[8];
	} VrsSid[] = 
	{
		{ SID_BEA01, "BEA01" },
		{ SID_BOOT2, "BOOT2" },
		{ SID_CD001, "CD001" },
		{ SID_CDW02, "CDW02" },
		{ SID_NSR02, "NSR02" },
		{ SID_NSR03, "NSR03" },
		{ SID_TEA01, "TEA01" },
	};

	blkNum    = this->GetBlkNum();
	blockSize = this->GetBlkSize();
	if ( (buffer = new BYTE [blockSize]) == NULL )
		return -1;
	try
	{
		VRS			*vrs;
		unsigned	n;

		n = 0;
		vrs = (VRS *)buffer;
		for(;;)
		{
			unsigned	adr;
			int			i, sidtype;

			adr = (32768 / blockSize) + n;	// 32KB offset
			if ( (INT64)adr >= blkNum )
				throw -1;
			if ( this->ReadSector( (INT64)adr, 1, blockSize, buffer) < 0 )
				throw -1;
			sidtype = -1;
			for ( i = 0; i < sizeof(VrsSid)/sizeof(*VrsSid); i++)
			{
				if ( strncmp((char *)vrs->sid, VrsSid[i].sid, 5) == 0 )
				{
					sidtype = VrsSid[i].type;
					break;
				}
			}
			if ( sidtype < 0 )
				break;
			vrsInfo.Add(sidtype, adr, vrs);
			switch ( sidtype )
			{
				case SID_BEA01:
					break;
				case SID_BOOT2:
					break;
				case SID_CD001:
					switch ( vrs->type )
					{
						case 0x00:
						case 0x01:
						case 0x02:
						case 0x03:
						case 0xFF:
						default  :
							break;
					}
					break;
				case SID_CDW02:
					break;
				case SID_NSR02:
				case SID_NSR03:
					break;
				case SID_TEA01:
					break;
			}
			n++;
		} // end of for(;;)
		retcode = n > 0 ? 0 : -1;
	}
	catch ( int ret ) { retcode = ret; }
	if ( buffer != NULL )
		delete [] buffer;

	return retcode;
}

//////////////////////////////////////////////////////////////////////////
//	Get Anchor Volume Descriptor Pointer

int		UdfLib::GetAVDP(UdfLib::ADVPInfo & avdpInfo)
{
	BYTE						*buffer = NULL;
	INT64						blkNum;
	unsigned					blockSize;
	int							retcode;

	blkNum    = this->GetBlkNum();
	blockSize = this->GetBlkSize();
	if ( (buffer = new BYTE [blockSize]) == NULL )
		return -1;
	try
	{
		static int		front_pos[] = { 256, 512 };
		static int		back_pos[] = { 0, 150, 256, 406 };
		int				i;

		avdpInfo.flag = 0;
		// check front tag
		for ( i = 0; i < sizeof(front_pos)/sizeof(*front_pos); i++)
		{
			unsigned	adr;
			AVDP		*avdp;

			adr = front_pos[i];
			if ( (INT64)adr >= blkNum )
				throw -1;
			if ( this->ReadSector((INT64)adr, 1, blockSize, buffer) < 0 )
				throw -1;
			avdp = (AVDP *)buffer;
			if ( this->CheckTag(avdp->tag, adr, TAGID_ANCHOR, 16, 496) < 0 )
				continue;
		//	fprintf(stdout, "Front AVDP OK: %u\n", adr);
			avdpInfo.p[0].loc = adr;
			memcpy( &avdpInfo.p[0].avdp, avdp, sizeof(AVDP));
			avdpInfo.flag |= 1;
			break;
		}
		// check back tag
		for ( i = 0; i < sizeof(back_pos)/sizeof(*back_pos); i++)
		{
			unsigned	adr;
			AVDP		*avdp;

			adr = (unsigned)(blkNum - 1 - back_pos[i]);
			if ( (INT64)adr >= blkNum )
				throw -1;
		//	fprintf(stdout, "Back AVDP Read %I64u - 1 - %u = %u\n", (INT64)blkNum, back_pos[i], adr);
			if ( this->ReadSector((INT64)adr, 1, blockSize, buffer) < 0 )
				throw -1;
			avdp = (AVDP *)buffer;
			if ( this->CheckTag(avdp->tag, adr, TAGID_ANCHOR, 16, 496) < 0 )
				continue;
		//	fprintf(stdout, "Back AVDP OK: %u\n", adr);
			avdpInfo.p[1].loc = adr;
			memcpy( &avdpInfo.p[1].avdp, avdp, sizeof(AVDP));
			avdpInfo.flag |= 2;
			break;
		}
		if ( (avdpInfo.flag & 3) == 1 )
		{
		//	fprintf(stdout,"Back AVDP error!\n");
			memcpy(&avdpInfo.p[1].avdp, &avdpInfo.p[0].avdp, sizeof(avdpInfo.p[0].avdp));
		}

		retcode = 0;
	}
	catch ( int ret ) { retcode = ret; }
	if ( buffer != NULL )
		delete [] buffer;

	return retcode;
}

//////////////////////////////////////////////////////////////////////////
//	Get Volume Descriptor Sequence

int		UdfLib::GetVDS(UdfLib::VdsInfo & vdsInfo, UdfLib::AVDP & avdp)
{
	unsigned	blockSize = this->GetBlkSize();
	BYTE		*mbuffer = NULL, *rbuffer = NULL;
	int			retcode;

	try
	{
		unsigned	location, length, bufsiz;

		location = this->U_endian32(avdp.mainVDSAdr.location);
		length   = this->U_endian32(avdp.mainVDSAdr.length);
		bufsiz = blockSize * 5 > length ? blockSize * 5 : length;
	//	fprintf(stderr, "loc %u, len %u, buf %u\n", (unsigned)location, (unsigned)length, bufsiz);
		if ( (mbuffer = new BYTE [bufsiz]) == NULL )
			throw -1;
		if ( this->ReadVDS(vdsInfo.mVolDes, mbuffer, location, length) < 0 )
			throw -1;

		location = this->U_endian32(avdp.reserveVDSAdr.location);
		length   = this->U_endian32(avdp.reserveVDSAdr.length);
		bufsiz = blockSize  * 5 > length ? blockSize * 5 : length;
	//	fprintf(stderr, "loc %u, len %u, buf %u\n", (unsigned)location, (unsigned)length, bufsiz);
		if ( (rbuffer = new BYTE [bufsiz]) == NULL )
			throw -1;
		if ( this->ReadVDS(vdsInfo.rVolDes, rbuffer, location, length) < 0 )
			throw -1;

		retcode = 0;
	}
	catch ( int ret ) { retcode = ret; }
	if ( mbuffer != NULL )
		delete [] mbuffer;
	if ( rbuffer != NULL )
		delete [] rbuffer;
	
	return retcode;
}

int		UdfLib::ReadVDS(UdfLib::VolDesInfo & volDesInfo, Uint8 * buf, Uint32 location, Uint32 length)
{
	Uint8				*buffer = NULL;
	unsigned			blockSize = this->GetBlkSize();
	unsigned			i, j;
	DTag				*vdtag;
	PrimaryVolDes		*pvd, *pvdt;
	LogVolDesc			*lvd, *lvdt;
	ImpUseVolDesc		*iuvd, *iuvdt;
	UnallocSpDesHead	*usd, *usdt;
	PartDesc 			*pd , *pdt;
	Uint8				zerotest;

//	fprintf(stdout,"ReadVDS() %u %u\n", location, length);
	if ( (buffer = new Uint8[blockSize]) == NULL )
		return -1;
	memset(buf, 0, length);
	pvd   = (PrimaryVolDes    *) buf;
	lvd   = (LogVolDesc       *)(buf + blockSize);
	iuvd  = (ImpUseVolDesc    *)(buf + blockSize * 2);
	usd   = (UnallocSpDesHead *)(buf + blockSize * 3);
	pd    = (PartDesc         *)(buf + blockSize * 4);

	vdtag = (DTag             *) buffer;
	pvdt  = (PrimaryVolDes    *) buffer;
	lvdt  = (LogVolDesc       *) buffer;
	iuvdt = (ImpUseVolDesc    *) buffer;
	usdt  = (UnallocSpDesHead *) buffer;
	pdt   = (PartDesc         *) buffer;
	for ( i = 0; i < length / blockSize; i++ )
	{
	//	fprintf(stdout,"ReadSector %u \n", location + i);
		if ( this->ReadSector((INT64)(location + i), 1, blockSize, buffer) < 0 )
			throw -1;
		this->CheckTag(*vdtag, location + i, -1, 0, 496);
		zerotest = 0;
		for ( j = 0; j < blockSize; j++)
			zerotest |= buffer[j];
	//	fprintf(stdout, "  loc:%u, id:%04X\n", location + i, this->U_endian16(vdtag->tagID));
		switch ( this->U_endian16(vdtag->tagID) )
		{
			case 0:	// blank or zeroed sector
				i = location;	// break for(;;) loop
				break;

			case TAGID_PVD:
				if ( this->U_endian16(pvd->tag.tagID) )
				{	// PVD already exists.  Replace if OK
					if ( memcmp(pvd->volID      , pvdt->volID      , sizeof(pvd->volID)   )
					  || memcmp(pvd->volSetID   , pvdt->volSetID   , sizeof(pvd->volSetID))
					  || memcmp(&pvd->desCharSet, &pvdt->desCharSet, sizeof(charspec     )) )
					{	// id mismatch previous 
					
					} else
					{
						if ( this->U_endian32(pvdt->volDescSeqNum) > pvd->volDescSeqNum )
						{	// replace
							memcpy(pvd, pvdt, blockSize);
							volDesInfo.AddDes(VOLDES_PVD, location + i, 1, sizeof(*pvd), &pvd->tag);
						} else
						{	// not replace
						}
					}
				} else
				{	// 1st PVD
					memcpy(pvd, pvdt, blockSize);
					volDesInfo.AddDes(VOLDES_PVD, location + i, 1, sizeof(*pvd), &pvd->tag);
				}
				break;

			case TAGID_POINTER:
				location = this->U_endian32(((VolDescPtr *)buffer)->nextVDS.location);
				length   = this->U_endian32(((VolDescPtr *)buffer)->nextVDS.length);
				this->track_volspace(volDesInfo, location, length / blockSize, "Next VDS sequence");
				i = (unsigned)-1;
				break;

			case TAGID_IUVD:
				if ( this->CheckRegid(&iuvdt->implementationID, E_REGID_IUVD) < 0 )
				{	// non-UDF IUVD, skip
				} else
				{
					if ( this->U_endian16(iuvd->tag.tagID) )
					{
						if ( this->U_endian32(iuvdt->volDescSeqNum) > this->U_endian32(iuvd->volDescSeqNum) )
						{	// replace
							memcpy(iuvd, iuvdt, blockSize);
							volDesInfo.AddDes(VOLDES_IUVD, location + i, 1, sizeof(*iuvd), &iuvd->tag);
						} else
						{	// not replace
						}
					} else
					{	// 1st IUVD
						memcpy(iuvd, iuvdt, blockSize);
						volDesInfo.AddDes(VOLDES_IUVD, location + i, 1, sizeof(*iuvd), &iuvd->tag);
					}
				}
				break;

			case TAGID_PD:
				if ( this->U_endian16(pd->tag.tagID) )
				{
					if ( this->U_endian16(pd->partNumber) != this->U_endian16(pdt->partNumber) )
					{	// PD mismatch
					} else
					{
						if ( this->U_endian32(pdt->volDescSeqNum) > this->U_endian32(pd->volDescSeqNum) )
						{	// replace
							memcpy(pd, pdt, blockSize);
							volDesInfo.AddDes(VOLDES_PD, location + i, 1, sizeof(*pd), &pd->tag);
						} else
						{	// not repalce
						}
					}
				} else
				{	// 1st PD
					memcpy(pd, pdt, blockSize);
					volDesInfo.AddDes(VOLDES_PD, location + i, 1, sizeof(*pd), &pd->tag);
				}
				break;

			case TAGID_LVD:
				if ( this->U_endian16(lvd->tag.tagID) )
				{
					if ( memcmp(lvd->logVolID, lvdt->logVolID, 128)
					  || memcmp(&lvd->desCharSet, &lvdt->desCharSet, sizeof(charspec)))
					{	// LVD mismatch
					} else
					{
						if ( this->U_endian32(lvdt->volDescSeqNum) > this->U_endian32(lvd->volDescSeqNum))
						{	// replace
							memcpy(lvd, lvdt, blockSize);
							volDesInfo.AddDes(VOLDES_LVD, location + i, 1, sizeof(*lvd), &lvd->tag);
						} else
						{	// not replace
						}
					}
				} else
				{	// 1st LVD
					memcpy(lvd, lvdt, blockSize);
					volDesInfo.AddDes(VOLDES_LVD, location + i, 1, sizeof(*lvd), &lvd->tag);
				}
				break;

			case TAGID_USD:
				if ( this->U_endian16(usd->tag.tagID) )
				{
					if ( this->U_endian32(usdt->volDescSeqNum) > this->U_endian32(usd->volDescSeqNum) )
					{	// replace
						memcpy(usd, usdt, blockSize);
						volDesInfo.AddDes(VOLDES_USD, location + i, 1, sizeof(*usd), &usd->tag);
					} else
					{	// not replace
					}
				} else
				{	// 1st USD
					memcpy(usd, usdt, blockSize);
					volDesInfo.AddDes(VOLDES_USD, location + i, 1, sizeof(*usd), &usd->tag);
				}
				break;

			case TAGID_TERM_DESC:
				i = length;	// break for(;;) loop
				break;

			default:	// unknown tag
				break;
		} // end of switch
		
	} // end of for (;;)

	return 0;
}
